/*
    SSSD

    Extended NSS Responder Interface

    Authors:
        Sumit Bose <sbose@redhat.com>

    Copyright (C) 2017 Red Hat

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <errno.h>

#include <sys/param.h> /* for MIN() */

#include "sss_client/sss_cli.h"
#include "sss_client/nss_mc.h"
#include "sss_client/nss_common.h"
#include "sss_client/idmap/sss_nss_idmap.h"
#include "sss_client/idmap/sss_nss_idmap_private.h"

#ifndef discard_const
#define discard_const(ptr) ((void *)((uintptr_t)(ptr)))
#endif

struct sss_nss_ex_data {
    char *items[EX_DATA_END];
};

const char *sss_nss_get_extra_data_item(struct sss_nss_ex_data *ex,
                                    enum extra_data_item item)
{
    if (item < 0 || item >= EX_DATA_END) {
        return NULL;
    }

    return ex->items[item];
}

void sss_nss_free_ex_data(struct sss_nss_ex_data *ex)
{
    size_t i;

    if (ex != NULL) {
        for (i = 0; i < EX_DATA_END; i++) {
            free(ex->items[i]);
        }
        free(ex);
    }
}

static int parse_extra_data(uint32_t extra_data_len, char *extra_data_blob,
                            struct sss_nss_ex_data **_ex)
{
    size_t c;
    size_t i;
    size_t len;
    uint32_t dummy;
    uint32_t extra_data_count;
    const char *tmp;
    struct sss_nss_ex_data *ex;
    int ret;

    if (_ex == NULL) {
        /* no extra data requested */
        return 0;
    }

    ex = calloc(1, sizeof(struct sss_nss_ex_data));
    if (ex == NULL) {
        return ENOMEM;
    }

    if (extra_data_len == 0) {
        for (i = 0; i < EX_DATA_END; i++) {
            ex->items[i] = NULL;
        }
        *_ex = ex;
        return 0;
    }

    if (extra_data_len < 2 * sizeof(uint32_t)) {
        ret = EINVAL;
        goto done;
    }

    c = 0;
    SAFEALIGN_COPY_UINT32(&dummy, extra_data_blob + c, &c);
    if (dummy != extra_data_len) {
        ret = EINVAL;
        goto done;
    }

    SAFEALIGN_COPY_UINT32(&extra_data_count, extra_data_blob + c, &c);
    if (c >= extra_data_len) {
        ret = EINVAL;
        goto done;
    }

    for (i = 0; i < extra_data_count; i++) {
        if (c >= extra_data_len) {
            ret = EINVAL;
            goto done;
        }
        tmp = (const char *) extra_data_blob + c;

        len = strnlen(tmp, (extra_data_len - c));
        c += len + 1;

        /* ignore additional strings in extra data but make sure the extra
         * data blob is consistent by continuing the loop */
        if (i < EX_DATA_END) {
            ex->items[i] = strdup(tmp);
        }
    }

    if (c != extra_data_len) {
        ret = EINVAL;
        goto done;
    }

    *_ex = ex;
    ret = 0;

done:
    if (ret != 0) {
        sss_nss_free_ex_data(ex);
    }

    return ret;
}

struct sss_nss_initgr_rep {
    gid_t *groups;
    long int *ngroups;
    long int *start;
};

struct nss_input {
    union {
        const char *name;
        uid_t uid;
        gid_t gid;
    } input;
    struct sss_cli_req_data rd;
    enum sss_cli_command cmd;
    union {
        struct sss_nss_pw_rep pwrep;
        struct sss_nss_gr_rep grrep;
        struct sss_nss_initgr_rep initgrrep;
    } result;
};

errno_t sss_nss_mc_get_with_extra(struct nss_input *inp,
                                       char **extra_data,
                                       uint32_t *extra_data_len)
{
    switch(inp->cmd) {
    case SSS_NSS_GETPWNAM:
    case SSS_NSS_GETPWNAM_EX:
        return sss_nss_mc_getpwnam_with_extra(inp->input.name,
                                              strlen(inp->input.name),
                                              inp->result.pwrep.result,
                                              inp->result.pwrep.buffer,
                                              inp->result.pwrep.buflen,
                                              extra_data, extra_data_len);
        break;
    case SSS_NSS_GETPWUID:
    case SSS_NSS_GETPWUID_EX:
        return sss_nss_mc_getpwuid_with_extra(inp->input.uid,
                                              inp->result.pwrep.result,
                                              inp->result.pwrep.buffer,
                                              inp->result.pwrep.buflen,
                                              extra_data, extra_data_len);
        break;
    case SSS_NSS_GETGRNAM:
    case SSS_NSS_GETGRNAM_EX:
        return sss_nss_mc_getgrnam_with_extra(inp->input.name,
                                              strlen(inp->input.name),
                                              inp->result.grrep.result,
                                              inp->result.grrep.buffer,
                                              inp->result.grrep.buflen,
                                              extra_data, extra_data_len);
        break;
    case SSS_NSS_GETGRGID:
    case SSS_NSS_GETGRGID_EX:
        return sss_nss_mc_getgrgid_with_extra(inp->input.gid,
                                              inp->result.grrep.result,
                                              inp->result.grrep.buffer,
                                              inp->result.grrep.buflen,
                                              extra_data, extra_data_len);
        break;
    case SSS_NSS_INITGR:
    case SSS_NSS_INITGR_EX:
        return sss_nss_mc_initgroups_dyn(inp->input.name,
                                         strlen(inp->input.name),
                                         -1 /* currently ignored */,
                                         inp->result.initgrrep.start,
                                         inp->result.initgrrep.ngroups,
                                         &(inp->result.initgrrep.groups),
                                         *(inp->result.initgrrep.ngroups));
        break;
    default:
        return EINVAL;
    }
}

int sss_get_ex(struct nss_input *inp, uint32_t flags, unsigned int timeout,
               const char *domain, struct sss_nss_ex_data **extra_data)
{
    uint8_t *repbuf = NULL;
    size_t replen;
    size_t len;
    uint32_t num_results;
    int ret;
    int time_left;
    int errnop;
    uint32_t extra_data_len = 0;
    void *extra_data_start = NULL;
    char *extra_data_out = NULL;
    uint32_t *mc_extra_data_len = 0;
    char **mc_extra_data_out = NULL;
    size_t c;
    gid_t *new_groups;
    size_t idx;
    bool skip_mc = false;

    if ((flags & SSS_NSS_EX_FLAG_NO_CACHE) != 0) {
        skip_mc = true;
    }

    if (extra_data != NULL) {
        mc_extra_data_out = &extra_data_out;
        mc_extra_data_len = &extra_data_len;
    }

    if (!skip_mc) {
        ret = sss_nss_mc_get_with_extra(inp, mc_extra_data_out,
                                        mc_extra_data_len);
        switch (ret) {
        case 0:
            ret = parse_extra_data(extra_data_len, extra_data_out,
                                   extra_data);
            if (ret != 0) {
                goto out;
            }
            return 0;
        case ERANGE:
            return ERANGE;
        case ENOENT:
            /* fall through, we need to actively ask the parent
             * if no entry is found */
            break;
        default:
            /* if using the mmaped cache failed,
             * fall back to socket based comms */
            break;
        }
    }

    sss_nss_timedlock(timeout, &time_left);

    if (!skip_mc) {
        /* previous thread might already initialize entry in mmap cache */
        ret = sss_nss_mc_get_with_extra(inp, mc_extra_data_out,
                                        mc_extra_data_len);
        switch (ret) {
        case 0:
            ret = parse_extra_data(extra_data_len, extra_data_out,
                                   extra_data);
            if (ret != 0) {
                goto out;
            }
            ret = 0;
            goto out;
        case ERANGE:
            ret = ERANGE;
            goto out;
        case ENOENT:
            /* fall through, we need to actively ask the parent
             * if no entry is found */
            break;
        default:
            /* if using the mmaped cache failed,
             * fall back to socket based comms */
            break;
        }
    }

    // TODO:
    //nret = sss_nss_get_getgr_cache(NULL, gid, GETGR_GID,
    //                               &repbuf, &replen, errnop);
    ret = sss_nss_make_request_timeout(inp->cmd, &inp->rd, time_left,
                                       &repbuf, &replen, &errnop);
    if (ret != NSS_STATUS_SUCCESS) {
        ret = errnop != 0 ? errnop : EIO;
        goto out;
    }

    /* Get number of results from repbuf. */
    SAFEALIGN_COPY_UINT32(&num_results, repbuf, NULL);

    /* no results if not found */
    if (num_results == 0) {
        ret = ENOENT;
        goto out;
    }

    if (inp->cmd == SSS_NSS_INITGR || inp->cmd == SSS_NSS_INITGR_EX) {
        if (*(inp->result.initgrrep.ngroups) < num_results) {
            new_groups = realloc(inp->result.initgrrep.groups,
                                 num_results * sizeof(gid_t));
            if (new_groups == NULL) {
                ret = ENOMEM;
                goto out;
            }

            inp->result.initgrrep.groups = new_groups;
        }
        *(inp->result.initgrrep.ngroups) = num_results;

        idx = 2 * sizeof(uint32_t);
        for (c = 0; c < num_results; c++) {
            SAFEALIGN_COPY_UINT32(&(inp->result.initgrrep.groups[c]),
                                  repbuf + idx, &idx);
            *(inp->result.initgrrep.start) += 1;
        }

        ret = 0;
        goto out;
    }

    /* only 1 result is accepted for this function */
    if (num_results != 1) {
        ret = EBADMSG;
        goto out;
    }

    len = replen - 8;
    if (inp->cmd == SSS_NSS_GETPWNAM || inp->cmd == SSS_NSS_GETPWUID || inp->cmd == SSS_NSS_GETPWNAM_EX || inp->cmd == SSS_NSS_GETPWUID_EX) {
        ret = sss_nss_getpw_readrep(&(inp->result.pwrep), repbuf+8, &len);
    } else if (inp->cmd == SSS_NSS_GETGRNAM || inp->cmd == SSS_NSS_GETGRGID || inp->cmd == SSS_NSS_GETGRNAM_EX || inp->cmd == SSS_NSS_GETGRGID_EX) {
        ret = sss_nss_getgr_readrep(&(inp->result.grrep), repbuf+8, &len);
    } else {
        ret = EINVAL;
        goto out;
    }
    if (ret) {
            // TODO:
            //if (ret == ERANGE)
            //sss_nss_save_getgr_cache(NULL, gid, GETGR_GID, &repbuf, replen);
        goto out;
    }

    if (len == 0) {
        /* no extra data */
        ret = 0;
        goto out;
    }

    extra_data_start = repbuf + replen - len;
    SAFEALIGN_COPY_UINT32(&extra_data_len, extra_data_start, NULL);
    if (extra_data_len != len) {
        /* Inconsistent extra data */
        ret = EINVAL;
        goto out;
    }

    if (extra_data != NULL) {
        if (inp->cmd == SSS_NSS_GETPWNAM || inp->cmd == SSS_NSS_GETPWUID || inp->cmd == SSS_NSS_GETPWNAM_EX || inp->cmd == SSS_NSS_GETPWUID_EX) {
            extra_data_out = inp->result.pwrep.result->pw_shell + strlen(inp->result.pwrep.result->pw_shell) + 1;
            if ((extra_data_out - inp->result.pwrep.buffer) + extra_data_len > inp->result.pwrep.buflen) {
                return ERANGE;
            }
        } else if (inp->cmd == SSS_NSS_GETGRNAM || inp->cmd == SSS_NSS_GETGRGID || inp->cmd == SSS_NSS_GETGRNAM_EX || inp->cmd == SSS_NSS_GETGRGID_EX) {
            if (inp->result.grrep.result->gr_mem == NULL || inp->result.grrep.result->gr_mem[0] == NULL) {
                extra_data_out = inp->result.grrep.result->gr_passwd + strlen(inp->result.grrep.result->gr_passwd) + 1;
            } else {
                for (c = 0; inp->result.grrep.result->gr_mem[c] != NULL; c++);
                --c;
                extra_data_out = inp->result.grrep.result->gr_mem[c] + strlen(inp->result.grrep.result->gr_mem[c]) + 1;
            }
        }

        memcpy(extra_data_out, extra_data_start, extra_data_len);
        ret = parse_extra_data(extra_data_len, extra_data_out,
                               extra_data);
        if (ret != 0) {
            goto out;
        }
    }

out:
    free(repbuf);

    sss_nss_unlock();
    return ret;
}

static int make_name_flag_req_data(const char *name, uint32_t flags,
                                   struct sss_cli_req_data *rd)
{
    size_t len;
    size_t name_len;
    uint8_t *data;
    int ret;

    if (name == NULL) {
        return EINVAL;
    }

    ret = sss_strnlen(name, SSS_NAME_MAX, &name_len);
    if (ret != 0) {
        return ret;
    }
    name_len++;

    len = name_len + sizeof(uint32_t);
    data = malloc(len);
    if (data == NULL) {
        return ENOMEM;
    }

    memcpy(data, name, name_len);
    SAFEALIGN_COPY_UINT32(data + name_len, &flags, NULL);

    rd->len = len;
    rd->data = data;

    return 0;
}

int sss_nss_getpwnam_ex(const char *name, struct passwd *pwd,
                        char *buffer, size_t buflen, struct passwd **result,
                        uint32_t flags, unsigned int timeout,
                        const char *domain, struct sss_nss_ex_data **extra_data)
{
    int ret;
    struct nss_input inp = {
        .input.name = name,
        .cmd = SSS_NSS_GETPWNAM_EX,
        .result.pwrep.result = pwd,
        .result.pwrep.buffer = buffer,
        .result.pwrep.buflen = buflen};

    if (buffer == NULL || buflen == 0) {
        return ERANGE;
    }

    ret = make_name_flag_req_data(name, flags, &inp.rd);
    if (ret != 0) {
        return ret;
    }

    *result = NULL;

    ret = sss_get_ex(&inp, flags, timeout, domain, extra_data);
    free(discard_const(inp.rd.data));
    if (ret == 0) {
        *result = inp.result.pwrep.result;
    }
    return ret;
}

int sss_nss_getpwuid_ex(uid_t uid, struct passwd *pwd,
                        char *buffer, size_t buflen, struct passwd **result,
                        uint32_t flags, unsigned int timeout,
                        const char *domain, struct sss_nss_ex_data **extra_data)
{
    int ret;
    uint32_t req_data[2];
    struct nss_input inp = {
        .input.uid = uid,
        .cmd = SSS_NSS_GETPWUID_EX,
        .rd.len = 2 * sizeof(uint32_t),
        .rd.data = &req_data,
        .result.pwrep.result = pwd,
        .result.pwrep.buffer = buffer,
        .result.pwrep.buflen = buflen};

    if (buffer == NULL || buflen == 0) {
        return ERANGE;
    }

    SAFEALIGN_COPY_UINT32(&req_data[0], &uid, NULL);
    SAFEALIGN_COPY_UINT32(&req_data[1], &flags, NULL);
    *result = NULL;

    ret = sss_get_ex(&inp, flags, timeout, domain, extra_data);
    if (ret == 0) {
        *result = inp.result.pwrep.result;
    }
    return ret;
}

int sss_nss_getgrnam_ex(const char *name, struct group *grp,
                        char *buffer, size_t buflen, struct group **result,
                        uint32_t flags, unsigned int timeout,
                        const char *domain, struct sss_nss_ex_data **extra_data)
{
    int ret;
    struct nss_input inp = {
        .input.name = name,
        .cmd = SSS_NSS_GETGRNAM_EX,
        .result.grrep.result = grp,
        .result.grrep.buffer = buffer,
        .result.grrep.buflen = buflen};

    if (buffer == NULL || buflen == 0) {
        return ERANGE;
    }

    ret = make_name_flag_req_data(name, flags, &inp.rd);
    if (ret != 0) {
        return ret;
    }

    *result = NULL;

    ret = sss_get_ex(&inp, flags, timeout, domain, extra_data);
    free(discard_const(inp.rd.data));
    if (ret == 0) {
        *result = inp.result.grrep.result;
    }
    return ret;
}

int sss_nss_getgrgid_ex(gid_t gid, struct group *grp,
                        char *buffer, size_t buflen, struct group **result,
                        uint32_t flags, unsigned int timeout,
                        const char *domain, struct sss_nss_ex_data **extra_data)
{
    int ret;
    uint32_t req_data[2];
    struct nss_input inp = {
        .input.gid = gid,
        .cmd = SSS_NSS_GETGRGID_EX,
        .rd.len = 2 * sizeof(uint32_t),
        .rd.data = &req_data,
        .result.grrep.result = grp,
        .result.grrep.buffer = buffer,
        .result.grrep.buflen = buflen};

    if (buffer == NULL || buflen == 0) {
        return ERANGE;
    }

    SAFEALIGN_COPY_UINT32(&req_data[0], &gid, NULL);
    SAFEALIGN_COPY_UINT32(&req_data[1], &flags, NULL);
    *result = NULL;

    ret = sss_get_ex(&inp, flags, timeout, domain, extra_data);
    if (ret == 0) {
        *result = inp.result.grrep.result;
    }
    return ret;
}

int sss_nss_getgrouplist_ex(const char *name, gid_t *groups, int *ngroups,
                            uint32_t flags, unsigned int timeout,
                            const char *domain)
{
    int ret;
    gid_t *new_groups;
    long int new_ngroups;
    long int start = 0;
    struct nss_input inp = {
        .input.name = name,
        .cmd = SSS_NSS_INITGR_EX};

    if (groups == NULL || ngroups == NULL || *ngroups == 0) {
        errno = EINVAL;
        return -1;
    }

    ret = make_name_flag_req_data(name, flags, &inp.rd);
    if (ret != 0) {
        return -1;
    }

    new_ngroups = *ngroups;
    new_groups = malloc(new_ngroups * sizeof(gid_t));
    if (new_groups == NULL) {
        free(discard_const(inp.rd.data));
        errno = ENOMEM;
        return -1;
    }

    inp.result.initgrrep.groups = new_groups,
    inp.result.initgrrep.ngroups = &new_ngroups;
    inp.result.initgrrep.start = &start;


    ret = sss_get_ex(&inp, flags, timeout, domain, NULL);
    free(discard_const(inp.rd.data));
    if (ret == 0) {
        memcpy(groups, new_groups, MIN(*ngroups, start) * sizeof(gid_t));
    }
    free(new_groups);

    if (start > *ngroups) {
        errno = ERANGE;
    } else {
        errno = 0;
    }
    *ngroups = start;

    return (errno == 0) ? start : -1;
}
